import os
import shutil
import sys

import arrow
import jinja2
from PIL import Image, ImageOps

from . import model, r2
from .const import *
from .model import Gallery, Album, Picture, PictureData, AlbumData, SortBy, GalleryData

"""
【关于返回值】
本项目的返回值有时采用 (result, err) 的形式,
err 是 str|None, 有内容表示有错误, 空字符串或 None 表示没错误.
"""

loader = jinja2.FileSystemLoader(Templates_Path)
jinja_env = jinja2.Environment(
    loader=loader, autoescape=jinja2.select_autoescape()
)

# 将templates 文件夹内除了 tmplfile 之外的全部文件都复制到 output 文件夹
tmplfile = dict(
    gallery_toml     = Gallery_Toml,
    album_toml       = Album_Toml,
    picture_toml     = Picture_Toml,
    index_html       = "index.html",
    album_index_html = "album_index.html",
    pic_html         = "pic.html",
    pics_id_list_js  = Pics_Id_List_JS,
)


def render_write(tmpl_name:str, output_path:Path, data:dict):
    tmpl = jinja_env.get_template(tmplfile[tmpl_name])
    rendered = tmpl.render(data)
    print(f"render and write {output_path}")
    output_path.write_text(rendered, encoding="utf-8")


def render_gallery_toml(gallery:Gallery):
    render_write('gallery_toml', Gallery_Toml_Path, dict(data=gallery))


def render_album_toml(album:Album):
    toml_path = CWD.joinpath(album.foldername, Album_Toml)
    render_write('album_toml', toml_path, dict(data=album))


def render_picture_toml(toml_path:Path, pic:Picture):
    render_write('picture_toml', toml_path, dict(data=pic))


def folder_not_empty(folder):
    return True if os.listdir(folder) else False


def init_gallery():
    """在一个空文件夹中初始化一个图库.

    :return: 有错返回 err: str, 无错返回 falsy 值。
    """
    if folder_not_empty(CWD):
        return f"Folder Not Empty: {CWD}"

    Output_Local_Path.mkdir()
    Output_Web_Path.mkdir()
    Output_R2_Path.mkdir()
    copy_templates()
    render_gallery_toml(Gallery.default(CWD.name))
    r2.add_to_r2_files([Index_HTML])
    print("图库创建成功。")
    print(f"请用文本编辑器打开 {Gallery_Toml_Path} 填写图库相关信息。")
    return None


def get_gallery(ctx):
    if is_gallery_folder():
        gallery = Gallery.loads()
        return gallery

    print("请先进入图库根目录, 或使用 'r2g init' 命令新建图库")
    ctx.exit()


def is_gallery_folder():
    """检查当前文件夹是否图库根目录"""
    return Templates_Path.exists() \
        and Output_Web_Path.exists() \
        and Output_Web_Path.exists() \
        and Gallery_Toml_Path.exists()


def copy_templates():
    src_folder = Path(__file__).parent.joinpath(Templates)
    shutil.copytree(src_folder, Templates_Path)


def create_album(name:str, gallery:Gallery):
    if err := model.check_filename(name):
        return err

    album_path = CWD.joinpath(name)
    if album_path.exists():
        return f"文件夹已存在: {name}"

    album_path.mkdir()
    album_toml_path = album_path.joinpath(Album_Toml)
    metadata_path = album_path.joinpath(Metadata)
    thumbs_path = album_path.joinpath(Thumbs)
    thumbs_path.mkdir()
    metadata_path.mkdir()
    album = Album.default(name)
    render_album_toml(album)
    gallery.add_album(name)
    render_gallery_toml(gallery)
    print("相册创建成功。")
    print(f"请用文本编辑器打开 {album_toml_path} 填写相册相关信息。")
    return None


def render_all(albums_pics:dict, gallery:Gallery, force=False):
    local_output_path = Output_Local_Path.joinpath(Index_HTML)
    web_output_path = Output_Web_Path.joinpath(Index_HTML)
    r2_output_path = Output_R2_Path.joinpath(Index_HTML)
    update_gallery = render_all_albums(albums_pics, gallery)
    if not force:
        force = update_gallery
    force = render_index_html(
        "local", "index_html", local_output_path, gallery, force)
    render_index_html(
        "web", "index_html", web_output_path, gallery, force)
    render_index_html(
        "r2", "index_html", r2_output_path, gallery, force)
    r2.add_to_r2_files([Index_HTML])


def update_all_albums(albums_pics:dict, gallery:Gallery):
    """返回 True 表示有错, 返回 False 表示无错."""
    for album, pics in albums_pics.items():
        if update_album(pics, Path(album), gallery):
            return True
    return False


def render_all_albums(
        albums_pics:dict,
        gallery:Gallery,
        force=False
) -> bool:
    """:return: True 表示需要重新渲染 gallery 首页."""
    update_gallery = False
    for album_folder, pics in albums_pics.items():
        album_path = Path(album_folder)
        album = Album.loads(album_path.joinpath(Album_Toml))
        album_data = album.to_data(album_path, gallery.bucket_url)
        pics_sorted = sort_pics(pics, album, album_path)

        local_album_folder = Output_Local_Path.joinpath(album.foldername)
        local_album_folder.mkdir(exist_ok=True)
        web_album_folder = Output_Web_Path.joinpath(album.foldername)
        web_album_folder.mkdir(exist_ok=True)
        r2_album_folder = Output_R2_Path.joinpath(album.foldername)
        r2_album_folder.mkdir(exist_ok=True)

        # 渲染相册内的图片
        pics_data_list = render_album_pics(
            pics_sorted,
            local_album_folder,
            web_album_folder,
            r2_album_folder,
            album_data,
            gallery.to_data()
        )

        # 渲染相册索引页
        update_gallery = render_album_index_html(
            local_album_folder,
            web_album_folder,
            r2_album_folder,
            gallery,
            album,
            album_data,
            pics_data_list,
            force=force,
        )

        # 添加待上传的文件 到 r2_files.json
        obj_names = [pic.r2_html for pic in pics_data_list]
        pics_js_name = f"{album.foldername}/{Pics_Id_List_JS}"
        obj_names.extend([album_data.r2_html, pics_js_name])
        r2.add_to_r2_files(obj_names)

    return update_gallery


def sort_pics(pics_paths:list[Path], album:Album, album_path:Path) -> list[Path]:
    sort_by = SortBy[album.sort_by]
    if sort_by is SortBy.List:
        return [album_path.joinpath(filename) for filename in album.pictures]

    pairs = pics_with_ctime(pics_paths)
    if sort_by is SortBy.CTimeDesc:
        pairs = sorted(pairs, key=lambda pair: pair[1], reverse=True)
    else:
        pairs = sorted(pairs, key=lambda pair: pair[1])

    return [pair[0] for pair in pairs]


def pics_with_ctime(pics_paths:list):
    """:return: (pic_path, ctime)"""
    pairs = []
    for pic_path in pics_paths:
        pic = Picture.loads(get_pic_toml_path(pic_path))
        pairs.append((pic_path, pic.ctime))
    return pairs


def resize_all_albums_pics(albums_pics:dict, gallery:Gallery):
    for pics in albums_pics.values():
        resize_oversize_pics(pics, gallery)


def check_all_albums_cover(albums_pics:dict):
    """:return: err:str | None"""
    for album_folder, pics in albums_pics.items():
        album = Album.loads(Path(album_folder).joinpath(Album_Toml))
        if not album.cover:
            return f"每个相册都必须指定封面, 请向相册 {album.foldername} 添加图片。" \
                   f"添加图片后执行 'r2g -update' 会自动指定封面。" \
                   f"若想手动指定封面, 可修改 '{album.foldername}/album.toml' 中的 cover 项目。"

        pics_name_list = [pic.name for pic in pics]
        if album.cover not in pics_name_list:
            return f"找不到封面图片: {album.cover}\n" \
                   f"请修改 '{album.foldername}/album.toml' 中的 cover, " \
                   f"确保 cover 指定的图片存在于相册文件夹 {album.foldername} 内。"
    return None


def check_all_double_names(albums_pics:dict):
    """
    :return: 零表示没有重复的文件名, 大于零表示有重复的文件名
    """
    albums = {}
    for album, pics in albums_pics.items():
        names = get_double_names(pics)
        if names:
            albums[Path(album).name] = names

    if len(albums) > 0:
        err = "请修改以下文件名, 使每个文件名都是唯一 (不分大小写, 不管后缀名)"
        print_bad_names(albums, err)

    return len(albums)


def check_all_bad_names(albums_pics:dict):
    """
    :return: 零表示没有不符合要求的文件名, 大于零表示有不符合要求的文件名
    """
    albums = {}
    for album, pics in albums_pics.items():
        names = get_bad_pic_names(pics)
        if names:
            albums[Path(album).name] = names

    if len(albums) > 0:
        err = "请修改以下文件名, \n" \
              "文件名只能使用 0-9, a-z, A-Z, _(下划线), -(短横线), .(点)\n" \
              "不能使用空格，请用下划线或短横线代替空格。\n" \
              "并且不可使用 'index' 作为文件名。\n"
        print_bad_names(albums, err)

    return len(albums)


def print_bad_names(albums:dict, err:str):
    print(err)
    for album_name, pic_names in albums.items():
        for pic_name in pic_names:
            print(f"{album_name}/{pic_name}")


def pic_paths_to_pic_data(album_folder:str, pic_paths:list[Path]) -> list[PictureData]:
    pic_data_list = []
    for pic_path in pic_paths:
        toml_path = get_pic_toml_path(pic_path)
        pic_data = Picture.loads(toml_path).to_data(album_folder, pic_path.name)
        pic_data_list.append(pic_data)
    return pic_data_list


def get_all_albums_pictures(gallery:Gallery):
    """获取全部相册的全部图片的 Path"""
    albums = get_all_albums(gallery)
    albums_pics = {}
    for album_path in albums:
        albums_pics[str(album_path)] = get_pic_files(album_path)
    return albums_pics


def get_pic_files(album_path:Path):
    """获取指定相册内全部图片的 Path"""
    files = album_path.glob("*.*")
    pics = [pic for pic in files if pic.is_file() and pic.name != Album_Toml]
    return pics


def get_all_albums(gallery:Gallery):
    """获取全部相册的 Path"""
    albums = []
    for album_name in gallery.albums:
        album_path = CWD.joinpath(album_name)
        if not album_path.exists():
            print(f"相册不存在: {album_name}")
            continue
        albums.append(album_path)
    return albums


def update_album(pics:list, album_path:Path, gallery:Gallery) -> bool:
    """返回 True 表示有错, 返回 False 表示无错."""
    oversize_pics = get_oversize_pics(pics, gallery)
    if oversize_pics:
        print_oversize_pics(oversize_pics)
        return True

    new_pics = []
    new_files_set = set()
    for pic_path in pics:
        img = open_image(pic_path)
        if img is None:
            print(f"Not Image: {pic_path.name}")
        else:
            if create_pic_toml_if_not_exists(img, pic_path):
                new_pics.append(pic_path.name)
                thumb_path = create_thumb_if_not_exists(img, pic_path, album_path, gallery)
                new_files_set.add(str(thumb_path))
                new_files_set.add(str(pic_path))

    r2.add_pics_to_r2_waiting(new_files_set)
    update_album_pictures(new_pics, album_path)
    return False


def update_album_pictures(new_pics:list[str], album_path:Path):
    """添加图片到相册. TODO: 删除图片."""
    if not new_pics:
        return
    album_toml_path = album_path.joinpath(Album_Toml)
    album = Album.loads(album_toml_path)
    album.pictures = new_pics + album.pictures
    if not album.cover:
        album.cover = new_pics[0]
    render_album_toml(album)


def get_double_names(pics:list) -> list[str]:
    """找出同一相册内的同名图片 (不分大小写, 不管后缀名)"""
    count = {}
    for pic in pics:
        stem = pic.stem.lower()
        if stem in count:
            item = count[stem]
            item[0] += 1
            item[1].append(pic.name)
        else:
            count[stem] = [1, [pic.name]]

    names = []
    for item in count.values():
        if item[0] > 1:
            names.extend(item[1])

    return names


def get_bad_pic_names(files:list[Path]) -> list[str]:
    """找出相册内不符合要求的图片文件名"""
    bad_id_list = ["index"]
    bad_names = []
    for file in files:
        if file.stem.lower() in bad_id_list:
            bad_names.append(file.name)
        if model.check_filename(file.name):
            bad_names.append(file.name)
    return bad_names


def resize_oversize_pics(pics:list, gallery:Gallery):
    oversize_pics = get_oversize_pics(pics, gallery)
    for pic in oversize_pics:
        img = open_image(pic)
        exif = img.getexif()
        img = resize_image(img, gallery)
        pic_path = pic.with_suffix(gallery.thumb_suffix())
        img.save(pic_path, gallery.image_output_format, exif=reset_exif(exif))
        print(f"Resize to {pic_path}")


def reset_exif(exif):
    """resize后, 图片的宽, 高, 方向可能发生变化"""
    exif[Orientation] = 1
    if ImageWidth in exif:
        del exif[ImageWidth]
    if ImageLength in exif:
        del exif[ImageLength]
    return exif


def print_oversize_pics(oversize_pics):
    if oversize_pics:
        print("以下图片体积超过上限，请手动缩小图片，或使用 `r2g --force-resize` 自动缩小图片。")
        print(f"另外，可在 {Gallery_Toml_Path} 文件中修改图片体积上限（包括高度、宽度、占用空间）。")
        for pic in oversize_pics:
            print(pic)


def get_oversize_pics(pics:list[Path], gallery:Gallery) -> list:
    oversize_pics = []
    for pic in pics:
        img = open_image(pic)
        if img is None:
            continue
        if is_image_oversize(img, pic, gallery):
            oversize_pics.append(pic)
    return oversize_pics


def get_image_datetime(img:Image):
    exif = img.getexif()
    if DateTimeOriginal in exif:
        dt = arrow.get(exif[DateTimeOriginal], ImageDateTimeFormat)
        return dt.to("local").format(model.RFC3339)
    if DateTime in exif:
        dt = arrow.get(exif[DateTime], ImageDateTimeFormat)
        return dt.to("local").format(model.RFC3339)
    return model.now()


def get_pic_toml_path(pic_path:Path) -> Path:
    pic_toml_name = pic_path.with_suffix(Dot_Toml).name.lower()
    return pic_path.parent.joinpath(Metadata, pic_toml_name)


def create_pic_toml_if_not_exists(img:Image, pic_path:Path):
    """
    :return: True 表示这是新图片, 否则返回 False
    """
    pic_toml_path = get_pic_toml_path(pic_path)
    if not pic_toml_path.exists():
        pic = Picture.default(pic_path, get_image_datetime(img))
        render_picture_toml(pic_toml_path, pic)
        return True
    return False


def create_thumb_if_not_exists(
        img:Image, pic_path:Path, album_path:Path, gallery:Gallery) -> Path:
    thumb_name = pic_path.with_suffix(gallery.thumb_suffix()).name.lower()
    thumb_path = album_path.joinpath(Thumbs, thumb_name)
    create_thumb(img, thumb_path, gallery)
    return thumb_path


def create_thumb(img:Image, thumb_path:Path, gallery:Gallery):
    img = ImageOps.exif_transpose(img)
    img = ImageOps.fit(img, gallery.thumbnail_size())
    img.save(thumb_path, gallery.image_output_format)
    print(f"Create thumbnail {thumb_path}")


def resize_image(img:Image, gallery:Gallery):
    """
    :return: Image | None
    """
    img = ImageOps.exif_transpose(img)
    changed = False
    width, height = float(img.size[0]), float(img.size[1])
    width_max, height_max = float(gallery.image_width_max), float(gallery.image_height_max)
    if height > height_max:
        ratio = height_max / height
        width = width * ratio
        height = height_max
        changed = True
    if width > width_max:
        ratio = width_max / width
        height = height * ratio
        width = width_max
        changed = True
    if changed:
        size = round(width), round(height)
        img = img.resize(size)
    return img


def is_image_oversize(img:Image, file:Path, gallery:Gallery):
    filesize = file.lstat().st_size
    width, height = img.size
    if width > gallery.image_width_max \
            or height > gallery.image_height_max \
            or filesize > gallery.image_size_max * MB:
        return True
    return False


def open_image(file):
    """
    :return: Image | None
    """
    try:
        img = Image.open(file)
    except OSError:
        img = None
    return img


def render_index_html(
        output_type:str,
        tmpl_name:str,
        output_path:Path,
        gallery:Gallery,
        force=False
) -> bool:
    """返回 True 表示执行了渲染。"""
    checksum = gallery.make_checksum()
    if gallery.checksum != checksum:
        gallery.checksum = checksum
        render_gallery_toml(gallery)
        force = True

    if force:
        render_write(tmpl_name, output_path, dict(
            output_type=output_type,
            gallery=gallery.to_data(),
            albums=gallery.get_albumdata(),
        ))

    return force


def render_album_index_html(
        local_output_folder:Path,
        web_output_folder:Path,
        r2_output_folder:Path,
        gallery:Gallery,
        album:Album,
        album_data:AlbumData,
        pics_data_list:list[PictureData],
        force=False
) -> bool:
    """:return: True 表示需要重新渲染 gallery 首页."""
    update_gallery = False
    checksum = album.make_checksum()
    if album.checksum != checksum:
        album.checksum = checksum
        render_album_toml(album)
        force = True
        update_gallery = True

    if force:
        local_output_path = local_output_folder.joinpath(Index_HTML)
        web_output_path = web_output_folder.joinpath(Index_HTML)
        r2_output_path = r2_output_folder.joinpath(Index_HTML)
        local_js_output_path = local_output_folder.joinpath(Pics_Id_List_JS)
        web_js_output_path = web_output_folder.joinpath(Pics_Id_List_JS)
        r2_js_output_path = r2_output_folder.joinpath(Pics_Id_List_JS)
        gallery_data = gallery.to_data()
        render_write("album_index_html", local_output_path, dict(
            gallery=gallery_data,
            album=album_data,
            pictures=pics_data_list,
            parent_dir=f"../../{album_data.name}/"
        ))
        render_write("album_index_html", web_output_path, dict(
            gallery=gallery_data,
            album=album_data,
            pictures=pics_data_list,
            parent_dir=f"{gallery.bucket_url}{album_data.name}/"
        ))
        render_write("album_index_html", r2_output_path, dict(
            gallery=gallery_data,
            album=album_data,
            pictures=pics_data_list,
        ))
        pics_id_list = [pic.file_id for pic in pics_data_list]
        render_write("pics_id_list_js", local_js_output_path, dict(
            pics=pics_id_list
        ))
        render_write("pics_id_list_js", web_js_output_path, dict(
            pics=pics_id_list
        ))
        render_write("pics_id_list_js", r2_js_output_path, dict(
            pics=pics_id_list
        ))

    return update_gallery


def render_album_pics(
        pics_paths:list[Path],
        local_album_folder:Path,
        web_album_folder:Path,
        r2_album_folder:Path,
        album:AlbumData,
        gallery:GalleryData,
        force=False
) -> list[PictureData]:
    """返回 PictureData 有用."""
    pic_data_list = []
    for pic_path in pics_paths:
        pic_data = render_pic_html(
            pic_path, local_album_folder, web_album_folder, r2_album_folder,
            album, gallery, force)
        pic_data_list.append(pic_data)
    return pic_data_list


def render_pic_html(
        pic_path:Path,
        local_album_folder:Path,
        web_album_folder:Path,
        r2_album_folder:Path,
        album:AlbumData,
        gallery:GalleryData,
        force=False
) -> PictureData:
    """返回 PictureData 有用."""
    pic_toml_path = get_pic_toml_path(pic_path)
    pic = Picture.loads(pic_toml_path)
    pic_data = pic.to_data(album.name, pic_path.name)
    checksum = pic.make_checksum()
    if pic.checksum != checksum:
        pic.checksum = checksum
        render_picture_toml(pic_toml_path, pic)
        force = True

    if force:
        pic_filename = f"{pic_data.file_id}{Dot_HTML}"
        local_output_path = local_album_folder.joinpath(pic_filename)
        web_output_path = web_album_folder.joinpath(pic_filename)
        r2_output_path = r2_album_folder.joinpath(pic_filename)
        render_write("pic_html", local_output_path, dict(
            pic=pic_data,
            album=album,
            gallery=gallery,
            parent_dir=f"../../{album.name}/"
        ))
        render_write("pic_html", web_output_path, dict(
            pic=pic_data,
            album=album,
            gallery=gallery,
            parent_dir=f"{gallery.bucket_url}{album.name}/"
        ))
        render_write("pic_html", r2_output_path,
                     dict(pic=pic_data, album=album, gallery=gallery))

    return pic_data


def rename_pic(name:str, pic_path:Path):
    pass


def set_use_proxy(sw:str, gallery:Gallery):
    use_proxy = False
    if sw.lower() in ["1", "on", "true"]:
        use_proxy = True

    gallery.use_proxy = use_proxy
    render_gallery_toml(gallery)
    print(f"设置成功\nuse proxy = {use_proxy}\nhttp proxy = {gallery.http_proxy}")

    if not gallery.http_proxy and use_proxy:
        print(f"未设置 proxy, 请用文本编辑器打开 {Gallery_Toml_Path} 填写 http proxy")



def print_err(err:str):
    """如果有错误就打印, 没错误就忽略."""
    if err:
        print(f"Error: {err}", file=sys.stderr)


def print_err_exist(ctx, err:str):
    """若有错误则打印并结束程序, 无错误则忽略."""
    if err:
        print(f"Error: {err}", file=sys.stderr)
        ctx.exit()



import hashlib
import json
from pathlib import Path
from typing import Iterable

import boto3
from botocore.config import Config
from botocore.exceptions import BotoCoreError

from .const import R2_Files_JSON_Path, R2_Waiting_JSON_Path, Thumbs, Output_R2_Path


def get_bucket(cfg):
    s3 = get_s3(cfg)
    return s3.Bucket(cfg.bucket_name)


def get_s3(cfg):
    return boto3.resource(
        's3',
        endpoint_url=cfg.endpoint_url,
        aws_access_key_id=cfg.aws_access_key_id,
        aws_secret_access_key=cfg.aws_secret_access_key,
        config=Config(proxies=get_proxies(cfg)),
    )


def get_proxies(cfg):
    if cfg.use_proxy:
        return dict(http=cfg.http_proxy, https=cfg.http_proxy)
    return None


def add_pics_to_r2_waiting(new_pics:Iterable):
    waiting = get_r2_waiting()
    waiting["upload"] = waiting["upload"].union(new_pics)
    write_r2_waiting(waiting)


def write_r2_waiting(waiting:dict):
    waiting["upload"] = list(waiting["upload"])
    waiting["delete"] = list(waiting["delete"])
    R2_Waiting_JSON_Path.write_text(json.dumps(waiting))


def get_r2_waiting() -> dict:
    if R2_Waiting_JSON_Path.exists():
        waiting = json.loads(R2_Waiting_JSON_Path.read_text())
        waiting["upload"] = set(waiting["upload"])
        waiting["delete"] = set(waiting["delete"])
        return waiting
    return dict(upload=set(), delete=set())


def write_r2_files_json(r2_files:dict):
    data = json.dumps(r2_files)
    R2_Files_JSON_Path.write_text(data)


def get_r2_files() -> dict:
    if R2_Files_JSON_Path.exists():
        return json.loads(R2_Files_JSON_Path.read_text())
    return {}


def add_to_r2_files(obj_names: list[str]):
    """obj_names 是 obj_name 的列表.

    将待上传的 obj_name 添加到 r2_files, checksum 是空字符串。
    注意: 待上传的 obj_name 对应的 checksum 总是空字符串,
    只有当上传时, 才更新真实的 checksum.
    """
    r2_files = get_r2_files()
    for obj_name in obj_names:
        r2_files[obj_name] = ""
    write_r2_files_json(r2_files)


def update_r2_files(new_r2_files:dict):
    old_r2_files = get_r2_files()
    for obj_name, checksum in new_r2_files.items():
        old_r2_files[obj_name] = checksum
    write_r2_files_json(old_r2_files)


def upload_file(file:str, obj_name:str, bucket) -> bool:
    """返回 False 表示上传失败。"""
    success = True
    try:
        bucket.upload_file(file, obj_name)
    except BotoCoreError as err:
        print(err)
        success = False
    return success


def upload_pic(pic_path:str, bucket):
    """返回 False 表示上传失败。"""
    parts = Path(pic_path).parts
    if parts[-2] == Thumbs:
        obj_name = "/".join(parts[-3:])
    else:
        obj_name = "/".join(parts[-2:])
    return upload_file(pic_path, obj_name, bucket)


def upload_pics(bucket):
    """上传图片及其缩略图到 Cloudflare R2"""
    r2_waiting = get_r2_waiting()
    success = set()
    for pic_path in r2_waiting["upload"]:
        print(f"upload -> {pic_path}")
        if upload_pic(pic_path, bucket):
            success.add(pic_path)
        else:
            print(f"上传失败: {pic_path}")
            break
    r2_waiting["upload"] = r2_waiting["upload"].difference(success)
    write_r2_waiting(r2_waiting)


def upload_assets(bucket):
    """上传 HTML/CSS 等文件到 Cloudflare R2"""
    r2_files = get_r2_files()
    success = {}
    for obj_name in r2_files:
        filepath = Output_R2_Path.joinpath(obj_name)
        checksum = file_checksum(filepath)
        if r2_files[obj_name] == checksum:
            continue

        print(f"upload -> {filepath}")
        if upload_file(str(filepath), obj_name, bucket):
            success[obj_name] =checksum
        else:
            print(f"上传失败: {filepath}")
    update_r2_files(success)


def file_checksum(filepath:Path) -> str:
    text = filepath.read_text()
    return hashlib.sha1(text.encode()).hexdigest()



from pathlib import Path

# 这里主要是一些字符串的常量
# 定义字符串常量后, 在写代码时就有自动补全, 很方便, 也能减少 typo

Templates        = "templates"
Output_Local     = "output_local"
Output_Web       = "output_web"
Output_R2        = "output_r2"
Gallery_Toml     = "gallery.toml"
Album_Toml       = "album.toml"
Picture_Toml     = "picture.toml"
Metadata         = "metadata"
Thumbs           = "thumbs"
Use_Proxy        = "use_proxy"
Http_Proxy       = "http_proxy"
R2_Files_JSON    = "r2_files.json"
R2_Waiting_JSON  = "r2_waiting.json"

Index_HTML       = "index.html"
Pics_Id_List_JS  = "pics_id_list.js"

Dot_HTML = ".html"
Dot_Toml = ".toml"
Dot_JPEG = ".jpeg"
# Dot_Webp = ".webp"

DateTimeOriginal = 36867
DateTime         = 306
Orientation      = 274
ImageWidth       = 256
ImageLength      = 257

RFC3339 = "YYYY-MM-DD HH:mm:ssZZ"
ImageDateTimeFormat = "YYYY:MM:DD HH:mm:ss"

MB = 1024 * 1024

CWD                  = Path.cwd().resolve()
Templates_Path       = CWD.joinpath(Templates)
Output_Local_Path    = CWD.joinpath(Output_Local)
Output_Web_Path      = CWD.joinpath(Output_Web)
Output_R2_Path       = CWD.joinpath(Output_R2)
Gallery_Toml_Path    = CWD.joinpath(Gallery_Toml)
R2_Waiting_JSON_Path = CWD.joinpath(R2_Waiting_JSON)
R2_Files_JSON_Path   = CWD.joinpath(R2_Files_JSON)



import click

from . import __version__, util, r2
from .const import Gallery_Toml_Path
from .util import print_err, print_err_exist, get_gallery

"""
【关于返回值】
本项目的返回值有时采用 (result, err) 的形式,
err 是 str|None, 有内容表示有错误, 空字符串或 None 表示没错误.
"""

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


def show_info(ctx):
    gallery = get_gallery(ctx)
    title, err = gallery.title()
    print_err_exist(ctx, err)

    print()
    print(f"[r2g]     {__file__}")
    print(f"[version] {__version__}")
    print(f"[repo]    https://github.com/ahui2016/R2-Gallery")
    print()
    print(f"[Gallery] {title}")
    print(f"[Author]  {gallery.author}")
    print(f"[Albums]  {len(gallery.albums)}")
    print()
    print("[R2 Home Page]")
    print(gallery.r2_html_url())
    print()
    print(f"[Image Width Max    ] {gallery.image_width_max} px")
    print(f"[Image Height Max   ] {gallery.image_height_max} px")
    print(f"[Image Size Max     ] {gallery.image_size_max} MB")
    print(f"[Image Output Format] {gallery.image_output_format}")
    print(f"[Thumbnail Size     ] {gallery.thumb_size}")
    print()
    print(f"[use proxy] {gallery.use_proxy}")
    proxy = gallery.http_proxy
    if not proxy and gallery.use_proxy:
        proxy = f"\n未设置 proxy, 请用文本编辑器打开 '{Gallery_Toml_Path}' 填写 http proxy"
    print(f"[http proxy] {proxy}")


@click.group(invoke_without_command=True)
@click.help_option("-h", "--help")
@click.option(
    "info", "-i", "-info", "-v", "-V",
    is_flag=True,
    help="Show information about the gallery.",
)
@click.option(
    "-update",
    is_flag=True,
    help="Add or remove toml files of all pictures."
)
@click.option(
    "--force-resize",
    is_flag=True,
    help="Force resize all oversize pictures."
)
@click.option("--use-proxy", help="Set '1' or 'on' or 'true' to use proxy.")
@click.pass_context
def cli(ctx, info, update, force_resize, use_proxy):
    """R2-Gallery: 个人独立相册，采用 Cloudflare R2 作为图片储存。

    https://github.com/ahui2016/R2-Gallery/
    """
    if info:
        show_info(ctx)
        ctx.exit()

    if use_proxy:
        gallery = get_gallery(ctx)
        util.set_use_proxy(use_proxy, gallery)
        ctx.exit()

    if update:
        gallery = get_gallery(ctx)
        albums_pics = util.get_all_albums_pictures(gallery)
        if util.check_all_bad_names(albums_pics) > 0:
            ctx.exit()
        if util.check_all_double_names(albums_pics) > 0:
            ctx.exit()
        if util.update_all_albums(albums_pics, gallery):
            ctx.exit()

        err = util.check_all_albums_cover(albums_pics)
        print_err_exist(ctx, err)

        util.render_all(albums_pics, gallery)
        ctx.exit()

    if force_resize:
        click.confirm("注意，强制缩小图片有可能覆盖原图，请先备份原图。确认执行吗？", abort=True)
        gallery = get_gallery(ctx)
        albums_pics = util.get_all_albums_pictures(gallery)
        util.resize_all_albums_pics(albums_pics, gallery)
        ctx.exit()

    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
        ctx.exit()


# 以上是主命令
# ########## #
# 以下是子命令


@cli.command(context_settings=CONTEXT_SETTINGS, name="init")
@click.pass_context
def init_command(ctx):
    """Initialize your gallery.

    初始化图库。请在一个空文件夹内执行 'r2g init'。
    """
    err = util.init_gallery()
    print_err(err)
    ctx.exit()


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.option("name", "-new", help="Create a new album.")
@click.pass_context
def album(ctx, name):
    """Operations about albums.

    关于相册的操作。
    """
    gallery = get_gallery(ctx)

    if name:
        err = util.create_album(name, gallery)
        print_err(err)
        ctx.exit()

    click.echo(ctx.get_help())
    ctx.exit()


@cli.command(context_settings=CONTEXT_SETTINGS)
@click.option("all_files", "-all", is_flag=True, help="Upload pictures and assets.")
@click.option("-pics", is_flag=True, help="Upload pictures.")
@click.option("-assets", is_flag=True, help="Upload assets.")
@click.pass_context
def upload(ctx, all_files, pics, assets):
    """Upload pictures or static files.

    上传图片或 HTML/CSS 等文件。
    """
    gallery = get_gallery(ctx)

    if all_files:
        bucket = r2.get_bucket(gallery)
        r2.upload_pics(bucket)
        r2.upload_assets(bucket)
    elif pics:
        bucket = r2.get_bucket(gallery)
        r2.upload_pics(bucket)
    elif assets:
        bucket = r2.get_bucket(gallery)
        r2.upload_assets(bucket)
    else:
        click.echo(ctx.get_help())
    ctx.exit()



import hashlib
import re
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path

import arrow
import tomli
import mistune

from .const import Gallery_Toml_Path, RFC3339, Metadata, Dot_Toml, CWD, \
    Album_Toml, Dot_JPEG, Index_HTML, Dot_HTML, Thumbs

Filename_Forbid_Pattern = re.compile(r"[^._0-9a-zA-Z\-]")
"""文件名只能使用 0-9, a-z, A-Z, _(下划线), -(短横线), .(点)。"""

Short_Notes_Limit = 512
"""简单描述的长度上限，单位: UTF8字符"""

Title_Limit = 32
"""从简介中提取标题时, 标题的长度上限, 单位: UTF8字符"""


class SortBy(Enum):
    """相册内的图片的排序方法"""
    CTime     = auto()  # 按图片拍摄或发布日期排序
    CTimeDesc = auto()  # 按图片拍摄或发布日期排序(倒序)
    List      = auto()  # 用列表指定顺序


def sort_by_from(text:str):
    n = len(text)
    if n <= len("List"):
        return SortBy.List
    elif n >= len("CTimeDesc"):
        return SortBy.CTimeDesc
    else:
        return SortBy.CTime


class Frontpage(Enum):
    """图库/相册首页的展示方式"""
    Story  = auto()  # 展示简介, 故事及列表
    Single = auto()  # 只展示一张图片
    List   = auto()  # 只展示列表


# 注意! 本软件暂时只能使用 JPEG, 原本打算让用户自由选择格式,
# 但后来想暂时先保持简单, 以后看情况再考虑让用户选择.
class ImageFormat(Enum):
    """缩小图片或生成缩略图时, 输出的图片格式"""
    WebP = auto()
    JPEG = auto()


def now():
    return arrow.now().format(RFC3339)


@dataclass
class PictureData:
    """用于生成前端 HTML"""
    file_id       : str  # 相当于 Picture.file_id
    filename      : str  # 图片文件名, 包括后缀, 不包括文件夹
    title         : str  # 提取自 Picture.notes 的第一行
    notes         : str  # 提取自 Picture.notes (不含第一行)
    story         : str  # Picture.story 转换为 HTML
    ctime         : str  # 相当于 Picture.ctime
    r2_html       : str  # 图片页面的 R2 对象名
    r2_pic_name   : str  # 图片本身的 R2 对象名
    r2_thumb_name : str  # 缩略图的 R2 对象名


@dataclass
class Picture:
    file_id  : str  # 文件名作为 id (不含后缀名, 转小写)
    notes    : str  # 简单描述 (纯文本格式, 第一行是图片标题)
    story    : str  # 图片的故事 (Markdown 格式)
    ctime    : str  # 拍摄或创作或发布日期
    checksum : str  # sha1, 用来判断 notes/story 有无变更

    @classmethod
    def default(cls, file:Path, ctime:str=None):
        if ctime is None:
            ctime = now()
        file_id = file.stem.lower()
        return Picture(
            file_id=file_id,
            notes=file_id,
            story="",
            ctime=ctime,
            checksum="",
        )

    @classmethod
    def loads(cls, toml_path:Path):
        """Loads TOML to a Picture."""
        data = tomli_loads(toml_path)
        pic = Picture(**data)
        pic.notes = pic.notes.strip()
        pic.story = pic.story.strip()
        return pic

    def make_checksum(self):
        text = self.notes + self.story + self.ctime
        return text_checksum(text)

    def title(self):
        """图片允许没有标题/简介"""
        pic_title, _, _ = split_notes(self.notes)
        return pic_title

    def r2_html(self, album_folder:str) -> str:
        """R2 HTML object name"""
        return f"{album_folder}/{self.file_id}{Dot_HTML}"

    def to_data(self, album_folder:str, pic_name:str) -> PictureData:
        """pic_name 是图片文件名, 包括后缀, 不包括文件夹."""
        title, notes, _ = split_notes(self.notes)
        return PictureData(
            file_id=self.file_id,
            filename=pic_name,
            title=title,
            notes=notes,
            story=mistune.html(self.story),
            ctime=self.ctime,
            r2_html=self.r2_html(album_folder),
            r2_pic_name=f"{album_folder}/{pic_name}",
            r2_thumb_name=f"{album_folder}/{Thumbs}/{self.file_id}{Dot_JPEG}",
        )


@dataclass
class AlbumData:
    """用于生成前端 HTML"""
    name               : str  # 相当于 Album.foldername
    author             : str  # 相当于 Album.author
    title              : str  # 提取自 Album.notes 的第一行
    notes              : str  # 提取自 Album.notes (不含第一行)
    story              : str  # Album.story 转换为 HTML
    sort_by            : str  # 相当于 Album.sort_by
    r2_html            : str  # 相当于 Album.r2_html
    cover_thumb_r2     : str  # Album.cover 的缩略图文件名
    cover_thumb_web    : str  # Album.cover 的缩略图网址
    cover_thumb_local  : str  # Album.cover 的缩略图本地路径
    cover_title        : str  # 提取自 Album.cover 的 notes


@dataclass
class Album:
    foldername : str  # 相册文件夹名称 (只能使用半角英数, 建议尽量简短)
    author     : str  # 作者, 留空表示跟随图库作者
    notes      : str  # 相册简介 (纯文本格式, 第一行是相册标题)
    story      : str  # 相册的故事 (Markdown 格式)
    sort_by    : str  # 相册内照片排序, 可选择 SortBy 里的五种排序方式
    pictures   : list # 图片文件名列表
    cover      : str  # 封面 (指定一个图片文件名)
    frontpage  : str  # 默认 Frontpage.Story
    checksum   : str  # sha1, 用来判断相册首页 HTML 要不要更新

    @classmethod
    def default(cls, foldername):
        return Album(
            foldername=foldername,
            author="",
            notes=foldername,
            story="",
            sort_by=SortBy.CTimeDesc.name,
            pictures=[],
            cover="",
            frontpage=Frontpage.Story.name,
            checksum="",
        )

    @classmethod
    def loads(cls, toml_path:Path):
        """Loads TOML to an Album."""
        data = tomli_loads(toml_path)
        album = Album(**data)
        album.notes = album.notes.strip()
        album.story = album.story.strip()
        album.frontpage = album.frontpage.capitalize()
        album.sort_by = sort_by_from(album.sort_by).name
        return album

    def make_checksum(self):
        pictures = ''.join(self.pictures)
        text = self.author + self.notes + self.story + self.sort_by + pictures \
               + self.cover + self.frontpage
        return text_checksum(text)

    def r2_html(self) -> str:
        """R2 HTML object name"""
        return f"{self.foldername}/{Index_HTML}"

    def to_data(self, album_path:Path, bucket_url:str) -> AlbumData:
        title, notes, err = split_notes(self.notes)
        if err:
            title = self.foldername
        toml_name = Path(self.cover).with_suffix(Dot_Toml).name
        toml_path = album_path.joinpath(Metadata, toml_name)
        cover = Picture.loads(toml_path)
        cover_filename = cover.file_id+Dot_JPEG
        return AlbumData(
            name=self.foldername,
            author=self.author,
            title=title,
            notes=notes,
            story=mistune.html(self.story),
            sort_by=self.sort_by,
            r2_html=self.r2_html(),
            cover_thumb_r2=cover_filename,
            cover_thumb_web=f"{bucket_url}{self.foldername}/thumbs/{cover_filename}",
            cover_thumb_local=f"../{self.foldername}/thumbs/{cover_filename}",
            cover_title=cover.title(),
        )


@dataclass
class GalleryData:
    author       : str  # 相当于 Gallery.author
    title        : str  # 提取自 Gallery.notes 的第一行
    notes        : str  # 提取自 Gallery.notes (不含第一行)
    story        : str  # Gallery.story 转换为 HTML
    frontpage    : str  # 相当于 Gallery.frontpage
    bucket_url   : str  # 相当于 Gallery.bucket_url


@dataclass
class Gallery:
    author    : str   # 图库作者 (默认:佚名)
    notes     : str   # 图库简介 (纯文本格式, 第一行是图库标题)
    story     : str   # 图库的故事 (Markdown 格式)
    frontpage : str   # 可选择 Frontpage 里的三种展示方式
    albums    : list  # 相册列表
    checksum  : str   # sha1, 用来判断图库首页 HTML 要不要更新
    use_proxy : str   # 是否使用 http proxy
    http_proxy : str  # 默认 http://127.0.0.1:1081

    image_width_max     : int  # 图片宽度上限, 单位: 像素
    image_height_max    : int  # 图片高度上限, 单位: 像素
    image_size_max      : int  # 图片体积上限, 单位: MB
    image_output_format : str  # 缩小图片或生成缩略图时输出的图片格式
    thumb_size          : int  # 缩略图边长 (缩略图总是正方形)

    endpoint_url          : str  # 以下 5 项是 Cloudflare R2 信息
    aws_access_key_id     : str
    aws_secret_access_key : str
    bucket_name           : str
    bucket_url            : str

    @classmethod
    def default(cls, title:str):
        author = "佚名"
        return Gallery(
            author=author,
            notes=title,
            story="",
            frontpage=Frontpage.Story.name,
            albums=[],
            checksum="",
            use_proxy="",
            http_proxy="http://127.0.0.1:1081",
            image_width_max=1000,
            image_height_max=1000,
            image_size_max=2,
            image_output_format=ImageFormat.JPEG.name,
            thumb_size=128,
            endpoint_url='https://<accountid>.r2.cloudflarestorage.com',
            aws_access_key_id = '<access_key_id>',
            aws_secret_access_key = '<access_key_secret>',
            bucket_name = '<bucket_name>',
            bucket_url = '<bucket_url>',
        )

    @classmethod
    def loads(cls):
        """Loads the Gallery config from Gallery_Toml_Path."""
        data = tomli_loads(Gallery_Toml_Path)
        gallery = Gallery(**data)
        gallery.notes = gallery.notes.strip()
        gallery.story = gallery.story.strip()
        gallery.frontpage = gallery.frontpage.capitalize()
        bucket_url = gallery.bucket_url
        gallery.bucket_url = bucket_url.rstrip("/") + "/"
        return gallery

    def r2_html_url(self) -> str:
        """R2 首页的完整网址"""
        return f"{self.bucket_url}{Index_HTML}"

    def to_data(self) -> GalleryData:
        title, notes, err = split_notes(self.notes)
        if err:
            title = err
        return GalleryData(
            author=self.author,
            title=title,
            notes=notes,
            story=mistune.html(self.story),
            frontpage=self.frontpage,
            bucket_url=self.bucket_url,
        )

    def title(self):
        """
        :return: (result, err)
        """
        gallery_title, _, err = split_notes(self.notes)
        return gallery_title, err

    def make_checksum(self):
        albums = ''.join(self.albums)
        text = self.author + self.notes + self.story + self.frontpage + albums
        return text_checksum(text)

    def add_album(self, album_name:str):
        self.albums.insert(0, album_name)

    def thumbnail_size(self):
        """用于 PIL.Image.thumbnail(size)"""
        return self.thumb_size, self.thumb_size

    def thumb_suffix(self):
        return f".{self.image_output_format.lower()}"

    def get_albumdata(self):
        albums = []
        for album_name in self.albums:
            album_path = CWD.joinpath(album_name)
            album_toml_path = album_path.joinpath(Album_Toml)
            album = Album.loads(album_toml_path)
            albums.append(album.to_data(album_path, self.bucket_url))
        return albums


def text_checksum(text:str) -> str:
    return hashlib.sha1(text.encode()).hexdigest()


def tomli_loads(file) -> dict:
    """正确处理 utf-16"""
    with open(file, "rb") as f:
        text = f.read()
        try:
            text = text.decode()  # Default encoding is 'utf-8'.
        except UnicodeDecodeError:
            text = text.decode("utf-16").encode("utf-8").decode("utf-8")
        return tomli.loads(text)


def split_notes(text:str) -> (str, str, str|None):
    """
    :return: (标题, 简介, 错误)
    """
    title, notes = split_first_line(text)
    err = None
    if not title:
        err = "未填写notes"
    return title, notes, err


def split_first_line(text:str) -> (str, str):
    """将 text 分成两部分, 第一部分是第一行, 其余是第二部分.
    其中第一行还有长度限制.

    注意, 有可能返回空字符串."""
    parts = text.split("\n", maxsplit=1)
    if len(parts) < 2:
        parts.append("")
    head, tail = parts[0], parts[1].strip()
    if len(head) > Title_Limit:
        head = head[:Title_Limit]
    return head, tail


def check_filename(name: str):
    """
    :return: 有错返回 err: str, 无错返回空字符串。
    """
    if Filename_Forbid_Pattern.search(name) is None:
        return False
    else:
        return "文件名/文件夹名只能使用 0-9, a-z, A-Z, _(下划线), -(短横线), .(点)" \
               "\n注意：不能使用空格，请用下划线或短横线代替空格。"
